==> commerce_gc_client-8.x-1.6.tar.gz: commerce_gc_client/src/Controller/WebhookHandler.php <==
<?php

namespace Drupal\commerce_gc_client\Controller;

use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\HttpFoundation\Response;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_price\Price;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_gc_client\Event\GoCardlessEvents;
use Drupal\commerce_gc_client\Event\WebhookEvent;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Database\Driver\mysql\Connection;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Config\ConfigFactoryInterface;

/**
 * Controller for handling webhooks from GoCardless.com.
 */
class WebhookHandler extends ControllerBase {

  /**
   * The database driver connection.
   *
   * @var \Drupal\Core\Database\Driver\mysql\Connection
   */
  protected $db;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $moduleSettings;

  /**
   * The event dispatcher.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected $eventDispatcher;

  /**
   * The "commerce_log" entity type storage.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $logStorage;

  /**
   * The "commerce_payment" entity type storage.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $paymentStorage;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private $logger;

  /**
   * The configuration settings for the GoCardless Client payment gateway.
   *
   * @var array
   */
  private $paymentGatewaySettings;

  /**
   * Constructs the WebhookHandler object.
   *
   * @param \Drupal\Core\Database\Driver\mysql\Connection $connection
   *   The database driver connection.
   * @param \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $eventDispatcher
   *   The event dispatcher.
   */
  public function __construct(Connection $connection, ContainerAwareEventDispatcher $eventDispatcher) {
    $this->db = $connection;
    $this->eventDispatcher = $eventDispatcher;
    $this->moduleSettings = $this->config('commerce_gc_client.settings')->get();
    $this->logStorage = $this->entityTypeManager()->getStorage('commerce_log');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('event_dispatcher')
    );
  }

  /**
   * Capture and pre-process incoming webhooks from GoCardless.
   */
  public function webhook() {
    $headers = function_exists('getallheaders') ? getallheaders() : $this->getallheaders();
    $provided_signature = $headers['Webhook-Signature'];
    $calculated_signature = FALSE;
    strpos($headers['Origin'], 'sandbox') !== FALSE ? $mode = 'sandbox' : $mode = 'live';

    if ($secret = $this->moduleSettings['webhook_secret_' . $mode]) {
      $webhook = file_get_contents('php://input');
      $calculated_signature = hash_hmac("sha256", $webhook, $secret);
    }
    $hash_equals = function_exists('hash_equals') ? hash_equals($provided_signature, $calculated_signature) : $this->hashEquals($provided_signature, $calculated_signature);

    if ($hash_equals) {
      // Process the events.
      $data = json_decode($webhook, TRUE);
      foreach ($data['events'] as $event) {

        switch ($event['resource_type']) {

          case 'mandates':
            if ($order_id = $this->getOrderId($event['links']['mandate'])) {
              if ($order = Order::load($order_id)) {
                $this->mandates($order, $event);
              }
            }
            break;

          case 'subscriptions':
            if ($order_id = $this->getOrderId($event['links']['subscription'])) {
              if ($order = Order::load($order_id)) {
                $this->subscriptions($order, $event);
              }
            }
            break;

          case 'payments':
            $payment_id = $event['links']['payment'];
            $partner = new GoCardlessPartner();
            $result = $partner->api([
              'mode' => $mode,
              'endpoint' => 'payments',
              'action' => 'get',
              'id' => $payment_id,
            ]);
            if ($result->response->status_code == 200) {
              $resource = $result->response->body->payments;
              if ($order_id = $this->getOrderId($resource->links->mandate)) {
                if ($order = Order::load($order_id)) {
                  $this->payments($order, $event, $resource);
                }
              }
            }
            break;
        }

        if (isset($order)) {
          // Get payment gateway settings if not already set.
          if (!isset($this->paymentGatewaySettings)) {
            $payment_gateway_id = $order->payment_gateway->entity->Id();
            $this->paymentGatewaySettings = $this->config('commerce_payment.commerce_payment_gateway.' . $payment_gateway_id)->get();
          }

          // Adds a webhook event to the Drupal log if the feature is enabled
          // in the modules settings form.
          if ($this->paymentGatewaySettings['configuration']['log_webhook']) {
            if (!isset($this->logger)) {
              $this->logger = $this->getLogger('commerce_gc_client');
            }
            $this->logger->notice('<pre>GoCardless webhook: <br />' . print_r($event, TRUE) . '</pre>');
          }
        }

        // Dispatch an event so that other modules can respond to webhook.
        !isset($resource) ? $resource = NULL : NULL;
        $webhook_event = new WebhookEvent($event, $resource, $order_id);
        $this->eventDispatcher->dispatch(GoCardlessEvents::WEBHOOK, $webhook_event);
      }
    }
    else {
      $this->logger->warning('Webhook cannot be processed because the webhook secret is not set or is invalid.');
      return new Response('Forbidden.', 403, ['Content-Type' => 'text/html']);
    }
    return new Response();
  }

  /**
   * Processes 'mandates' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   */
  private function mandates($order, array $event) {
    $this->logStorage->generate($order, 'webhook_description', [
      'description' => $event['details']['description'],
    ])->save();

    $this->db->update('commerce_gc_client')->fields([
      'gc_mandate_status' => $event['action'],
    ])->condition('order_id', $order->id())->execute();
  }

  /**
   * Processes 'subscription' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   */
  private function subscriptions($order, array $event) {
    $this->logStorage->generate($order, 'webhook_description', [
      'description' => $event['details']['description'],
    ])->save();

    if (!in_array($event['action'], ['payment_created', 'amended'])) {
      $this->db->update('commerce_gc_client_item')->fields([
        'gc_subscription_status' => $event['action'],
      ])->condition('gc_subscription_id', $event['links']['subscription'])->execute();
    }
  }

  /**
   * Processes 'payments' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   * @param object $payment
   *   A GoCardless payment object that is retreived from via the API upon
   *   receipt of a 'payments' webhook.
   */
  private function payments($order, array $event, $payment) {
    if (!$this->paymentStorage) {
      $this->paymentStorage = $this->entityTypeManager()->getStorage('commerce_payment');
    }
    if ($event['action'] == 'created') {
      $this->paymentStorage->create([
        'state' => 'new',
        'amount' => new Price((string) ($payment->amount / 100), $payment->currency),
        'payment_gateway' => $order->payment_gateway->entity->Id(),
        'order_id' => $order->id(),
        'remote_id' => $payment->id,
        'remote_state' => $payment->status,
      ])->save();
    }
    else {
      $payment_id = $this->db->select('commerce_payment', 'p')->fields('p', ['payment_id'])->condition('remote_id', $payment->id)->execute()->fetchField();
      $payment = Payment::load($payment_id);

      switch (TRUE) {
        case in_array($event['action'], [
          'pending_customer_approval',
          'pending_submission',
          'submitted',
        ]):
          $payment->setState('authorization');
          break;

        case in_array($event['action'], [
          'confirmed',
          'paid_out',
        ]):
          $payment->setState('completed');
          break;

        case in_array($event['action'], [
          'cancelled',
          'customer_approval_denied',
          'failed',
          'charged_back',
        ]):
          $payment->setState('authorization_voided');
          break;
      }
      $payment->save();
    }
  }

  /**
   * Get Commerce order ID.
   *
   * @param string $id
   *   Either a GoCardless mandate ID or a GoCardless subscription ID.
   *
   * @return int
   *   The Commerce order ID for the mandate or subscription.
   */
  private function getOrderId($id) {
    $query = $this->db->select('commerce_gc_client', 'g');
    if (substr($id, 0, 2) == 'MD') {
      $query->condition('gc_mandate_id', $id);
    }
    else {
      $query->join('commerce_gc_client_item', 'i', 'g.gcid = i.gcid');
      $query->condition('gc_subscription_id', $id);
    }
    return $query->fields('g', ['order_id'])->execute()->fetchField();
  }

  /**
   * Timing attack safe string comparison.
   *
   * Compares two strings using the same time whether they're equal or not.
   * Required when hash_equals() function is not present (PHP < 5.6.0).
   *
   * @param string $known_string
   *   The string of known length to compare against.
   * @param string $user_string
   *   The user-supplied string.
   *
   * @return bool
   *   Returns TRUE when the two strings are equal, FALSE otherwise.
   */
  private function hashEquals($known_string, $user_string) {
    $ret = 0;
    if (strlen($known_string) !== strlen($user_string)) {
      $user_string = $known_string;
      $ret = 1;
    }
    $res = $known_string ^ $user_string;
    for ($i = strlen($res) - 1; $i >= 0; --$i) {
      $ret |= ord($res[$i]);
    }
    return !$ret;
  }

  /**
   * Fetch all HTTP request headers.
   *
   * Required for Nginx servers that do not support getallheaders().
   *
   * @return mixed
   *   An associative array of all the HTTP headers in the current request, or
   *   FALSE on failure.
   */
  private function getallheaders() {
    $headers = [];
    foreach ($_SERVER as $name => $value) {
      if (substr($name, 0, 5) == 'HTTP_') {
        $headers[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
      }
    }
    return $headers;
  }

}

==> commerce_gc_client-8.x-1.7.tar.gz: commerce_gc_client/src/Controller/WebhookHandler.php <==
<?php

namespace Drupal\commerce_gc_client\Controller;

use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\HttpFoundation\Response;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_price\Price;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_gc_client\Event\GoCardlessEvents;
use Drupal\commerce_gc_client\Event\WebhookEvent;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Database\Driver\mysql\Connection;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Config\ConfigFactoryInterface;

/**
 * Controller for handling webhooks from GoCardless.com.
 */
class WebhookHandler extends ControllerBase {

  /**
   * The database driver connection.
   *
   * @var \Drupal\Core\Database\Driver\mysql\Connection
   */
  protected $db;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $moduleSettings;

  /**
   * The event dispatcher.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected $eventDispatcher;

  /**
   * The "commerce_log" entity type storage.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $logStorage;

  /**
   * The "commerce_payment" entity type storage.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $paymentStorage;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private $logger;

  /**
   * The configuration settings for the GoCardless Client payment gateway.
   *
   * @var array
   */
  private $paymentGatewaySettings;

  /**
   * Constructs the WebhookHandler object.
   *
   * @param \Drupal\Core\Database\Driver\mysql\Connection $connection
   *   The database driver connection.
   * @param \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $eventDispatcher
   *   The event dispatcher.
   */
  public function __construct(Connection $connection, ContainerAwareEventDispatcher $eventDispatcher) {
    $this->db = $connection;
    $this->eventDispatcher = $eventDispatcher;
    $this->moduleSettings = $this->config('commerce_gc_client.settings')->get();
    $this->logStorage = $this->entityTypeManager()->getStorage('commerce_log');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('event_dispatcher')
    );
  }

  /**
   * Capture and pre-process incoming webhooks from GoCardless.
   */
  public function webhook() {
    $headers = function_exists('getallheaders') ? getallheaders() : $this->getallheaders();
    $provided_signature = $headers['Webhook-Signature'];
    $calculated_signature = FALSE;
    strpos($headers['Origin'], 'sandbox') !== FALSE ? $mode = 'sandbox' : $mode = 'live';

    if ($secret = $this->moduleSettings['webhook_secret_' . $mode]) {
      $webhook = file_get_contents('php://input');
      $calculated_signature = hash_hmac("sha256", $webhook, $secret);

      $hash_equals = function_exists('hash_equals') ? hash_equals($provided_signature, $calculated_signature) : $this->hashEquals($provided_signature, $calculated_signature);

      if ($hash_equals) {
        // Process the events.
        $data = json_decode($webhook, TRUE);
        foreach ($data['events'] as $event) {

          switch ($event['resource_type']) {

            case 'mandates':
              if ($order_id = $this->getOrderId($event['links']['mandate'])) {
                if ($order = Order::load($order_id)) {
                  $this->mandates($order, $event);
                }
              }
              break;

            case 'subscriptions':
              if ($order_id = $this->getOrderId($event['links']['subscription'])) {
                if ($order = Order::load($order_id)) {
                  $this->subscriptions($order, $event);
                }
              }
              break;

            case 'payments':
              $payment_id = $event['links']['payment'];
              $partner = new GoCardlessPartner();
              $result = $partner->api([
                'mode' => $mode,
                'endpoint' => 'payments',
                'action' => 'get',
                'id' => $payment_id,
              ]);
              if ($result->response->status_code == 200) {
                $resource = $result->response->body->payments;
                if ($order_id = $this->getOrderId($resource->links->mandate)) {
                  if ($order = Order::load($order_id)) {
                    $this->payments($order, $event, $resource);
                  }
                }
              }
              break;
          }

          if (isset($order)) {
            // Get payment gateway settings if not already set.
            if (!isset($this->paymentGatewaySettings)) {
              $payment_gateway_id = $order->payment_gateway->entity->Id();
              $this->paymentGatewaySettings = $this->config('commerce_payment.commerce_payment_gateway.' . $payment_gateway_id)->get();
            }

            // Adds a webhook event to the Drupal log if the feature is enabled
            // in the modules settings form.
            if ($this->paymentGatewaySettings['configuration']['log_webhook']) {
              if (!isset($this->logger)) {
                $this->logger = $this->getLogger('commerce_gc_client');
              }
              $this->logger->notice('<pre>GoCardless webhook: <br />' . print_r($event, TRUE) . '</pre>');
            }
          }

          // Dispatch an event so that other modules can respond to webhook.
          !isset($resource) ? $resource = NULL : NULL;
          $webhook_event = new WebhookEvent($event, $resource, $order_id);
          $this->eventDispatcher->dispatch(GoCardlessEvents::WEBHOOK, $webhook_event);
        }
      }
    }
    else {
      $this->logger = $this->getLogger('commerce_gc_client')->warning('Webhook cannot be processed because the webhook secret is not set or is invalid.');
      return new Response('Forbidden.', 403, ['Content-Type' => 'text/html']);
    }
    return new Response();
  }

  /**
   * Processes 'mandates' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   */
  private function mandates($order, array $event) {
    $this->logStorage->generate($order, 'webhook_description', [
      'description' => $event['details']['description'],
    ])->save();

    $this->db->update('commerce_gc_client')->fields([
      'gc_mandate_status' => $event['action'],
    ])->condition('order_id', $order->id())->execute();
  }

  /**
   * Processes 'subscription' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   */
  private function subscriptions($order, array $event) {
    $this->logStorage->generate($order, 'webhook_description', [
      'description' => $event['details']['description'],
    ])->save();

    if (!in_array($event['action'], ['payment_created', 'amended'])) {
      $this->db->update('commerce_gc_client_item')->fields([
        'gc_subscription_status' => $event['action'],
      ])->condition('gc_subscription_id', $event['links']['subscription'])->execute();
    }
  }

  /**
   * Processes 'payments' webhooks.
   *
   * @param object $order
   *   Commerce order entity.
   * @param array $event
   *   The webhook event as provided by GoCardless.
   * @param object $payment
   *   A GoCardless payment object that is retreived from via the API upon
   *   receipt of a 'payments' webhook.
   */
  private function payments($order, array $event, $payment) {
    if (!$this->paymentStorage) {
      $this->paymentStorage = $this->entityTypeManager()->getStorage('commerce_payment');
    }
    if ($event['action'] == 'created') {
      $this->paymentStorage->create([
        'state' => 'new',
        'amount' => new Price((string) ($payment->amount / 100), $payment->currency),
        'payment_gateway' => $order->payment_gateway->entity->Id(),
        'order_id' => $order->id(),
        'remote_id' => $payment->id,
        'remote_state' => $payment->status,
      ])->save();
    }
    else {
      $payment_id = $this->db->select('commerce_payment', 'p')->fields('p', ['payment_id'])->condition('remote_id', $payment->id)->execute()->fetchField();
      $payment = Payment::load($payment_id);

      switch (TRUE) {
        case in_array($event['action'], [
          'pending_customer_approval',
          'pending_submission',
          'submitted',
        ]):
          $payment->setState('authorization');
          break;

        case in_array($event['action'], [
          'confirmed',
          'paid_out',
        ]):
          $payment->setState('completed');
          break;

        case in_array($event['action'], [
          'cancelled',
          'customer_approval_denied',
          'failed',
          'charged_back',
        ]):
          $payment->setState('authorization_voided');
          break;
      }
      $payment->save();
    }
  }

  /**
   * Get Commerce order ID.
   *
   * @param string $id
   *   Either a GoCardless mandate ID or a GoCardless subscription ID.
   *
   * @return int
   *   The Commerce order ID for the mandate or subscription.
   */
  private function getOrderId($id) {
    $query = $this->db->select('commerce_gc_client', 'g');
    if (substr($id, 0, 2) == 'MD') {
      $query->condition('gc_mandate_id', $id);
    }
    else {
      $query->join('commerce_gc_client_item', 'i', 'g.gcid = i.gcid');
      $query->condition('gc_subscription_id', $id);
    }
    return $query->fields('g', ['order_id'])->execute()->fetchField();
  }

  /**
   * Timing attack safe string comparison.
   *
   * Compares two strings using the same time whether they're equal or not.
   * Required when hash_equals() function is not present (PHP < 5.6.0).
   *
   * @param string $known_string
   *   The string of known length to compare against.
   * @param string $user_string
   *   The user-supplied string.
   *
   * @return bool
   *   Returns TRUE when the two strings are equal, FALSE otherwise.
   */
  private function hashEquals($known_string, $user_string) {
    $ret = 0;
    if (strlen($known_string) !== strlen($user_string)) {
      $user_string = $known_string;
      $ret = 1;
    }
    $res = $known_string ^ $user_string;
    for ($i = strlen($res) - 1; $i >= 0; --$i) {
      $ret |= ord($res[$i]);
    }
    return !$ret;
  }

  /**
   * Fetch all HTTP request headers.
   *
   * Required for Nginx servers that do not support getallheaders().
   *
   * @return mixed
   *   An associative array of all the HTTP headers in the current request, or
   *   FALSE on failure.
   */
  private function getallheaders() {
    $headers = [];
    foreach ($_SERVER as $name => $value) {
      if (substr($name, 0, 5) == 'HTTP_') {
        $headers[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
      }
    }
    return $headers;
  }

}
